"""Contains the code related to parsing Cmus output."""

from collections import defaultdict
from functools import reduce, partial

from .constants import FIELDS
from .types import StatusInformation


def parse_status_information(informations):
    """Parse the status informations from the informations list.

    :param informations: The list containing the various song's information
    :type informations: str
    :returns: :class:`collections.defaultdict`
    """
    fields = set(FIELDS)
    status_information = defaultdict(list)
    current_status = None
    for word in informations.split():
        if word in fields:
            current_status = word
        if current_status and word not in fields:
            status_information[current_status].append(word)

    return _format_status_information_fields(
        status_information,
        partial(_format_artist_field, 'artist'),
        partial(_format_artist_field, 'albumartist'),
        _format_duration_field,
        partial(_format_integer_field, 'tracknumber'),
        partial(_format_integer_field, 'discnumber'),
        _format_status_field,
        _format_left_fields
    )


def _format_status_information_fields(status_information, *formatters):
    """Format the status informations dictionary.

    :param status_information: The various fields information
    :type status_information: :class:`collections.defaultdict`
    :param formatters: The formatters functions
    :type formatters: tuple<callable>
    :returns: A status information object containing the various information
    :rtype: :class:`StatusInformation`
    """
    return StatusInformation(
        **reduce(
            lambda information, formatter: formatter(information),
            formatters,
            status_information
        )
    )


def _format_artist_field(field_name, status_information):
    """Format the *artist* field.

    :param status_information: The various fields information
    :type status_information: :class:`collections.defaultdict`
    :returns: The updated status informations
    :rtype: :class:`collections.defaultdict`
    """
    status_information[field_name] = ' '.join(status_information[field_name])
    return status_information


def _format_duration_field(status_information):
    """Format the *duration* field.

    :param status_information: The various fields information
    :type status_information: :class:`collections.defaultdict`
    :returns: The updated status informations
    :rtype: :class:`collections.defaultdict`
    """
    try:
        duration = int(''.join(status_information['duration']))
        status_information['duration'] = '{0:02d}:{1:02d}'.format(
            duration // 60,
            duration % 60
        )
    except ValueError:
        status_information.pop('duration', None)

    return status_information


def _format_status_field(status_information):
    """Format the *status* field.

    :param status_information: The various fields information
    :type status_information: :class:`collections.defaultdict`
    :returns: The updated status informations
    :rtype: :class:`collections.defaultdict`
    """
    status_information['status'] = (' '.join(status_information['status'])
                                       .capitalize())
    return status_information


def _format_integer_field(field_name, status_information):
    """Format an integer field.

    :param field_name: The name of the field
    :type field_name: str
    :param status_information: The various fields information
    :type status_information: :class:`collections.defaultdict`
    :returns: The updated status informations
    :rtype: :class:`collections.defaultdict`
    """
    try:
        status_information[field_name] = int(''.join(status_information[field_name]))
    except ValueError:
        status_information.pop(field_name, None)

    return status_information


def _format_left_fields(status_information):
    """Format the unformatted fields.

    :param status_information: The various fields information
    :type status_information: :class:`collections.defaultdict`
    :returns: The updated status informations
    :rtype: :class:`collections.defaultdict`
    """
    for key, value in status_information.copy().items():
        if isinstance(value, list) and value:
            status_information[key] = ' '.join(value)
        elif not value:
            status_information.pop(key, None)
    return status_information



"""Contains the constants used by the library."""


#: Default text for the display values
DEFAULT_STATUS_DISPLAY = 'N/A'
#: The various field contained in the metadata
FIELDS = (
    'status',
    'file',
    'artist',
    'album',
    'discnumber',
    'tracknumber',
    'title',
    'date',
    'duration',
    'albumartist'
)
#: The icons by the media player status
ICONS_BY_STATUS = {
    'paused': 'stock_media-pause',
    'playing': 'stock_media-play',
    'stopped': 'stock_media-stop'
}
#: The icon's path to be used by default
DEFAULT_ICON_PATH = ''
#: The default timeout for the notification to disappear
DEFAULT_TIMEOUT = 5000
#: The default name of the application
DEFAULT_APPLICATION_NAME = 'Cmus'
#: The default message body of the notifications
DEFAULT_MESSAGE_BODY = ('<b>Title:</b> {title}\n'
                        '<b>Artist:</b> {artist}\n'
                        '<b>Album:</b> {album}\n'
                        '<b>Date:</b> {date}\n'
                        '<b>Track:</b> {tracknumber}\n'
                        '<b>Disc:</b> {discnumber}\n'
                        '<b>Duration:</b> {duration}')
#: The default message title of the notifications
DEFAULT_MESSAGE_TITLE = 'Current Status: {status}'
#: The default path of the configuration file
DEFAULT_CONFIGURATION_FILE = '~/.cmus-notify'
#: The various configuration sections
CONFIGURATION_SECTIONS = (
    'notifications',
    'format'
)
#: The various options
DEFAULT_OPTIONS = {
    'body': DEFAULT_MESSAGE_BODY,
    'title': DEFAULT_MESSAGE_TITLE,
    'application_name': DEFAULT_APPLICATION_NAME,
    'timeout': DEFAULT_TIMEOUT
}



"""Contains code to format the current status of Cmus."""

from .constants import DEFAULT_MESSAGE_BODY, DEFAULT_MESSAGE_TITLE


def format_notification_message(status_information, **kwargs):
    """Format the :class:`cmus_notify.types.StatusInformation` to be send.

    :param status_information: The information to be sent
    :type status_information: :class:`cmus_notify.types.StatusInformation`
    """
    title = (kwargs.get('title', DEFAULT_MESSAGE_TITLE)
                   .format(**status_information.__dict__))
    text = (kwargs.get('body', DEFAULT_MESSAGE_BODY)
                  .format(**status_information.__dict__))
    return title, text



"""Contains the code related to configuration file parsing."""

from configparser import ConfigParser

from .constants import CONFIGURATION_SECTIONS, DEFAULT_OPTIONS
from .options import parse_arguments


def parse_configuration_file():
    """Parse the configuration file. If there are any command-line arguments,
       override the configuration values with the one from the command line.

       :returns: The various options
       :rtype: dict
    """
    arguments = parse_arguments()
    configuration_parser = ConfigParser()
    options = {}
    if configuration_parser.read(arguments.configuration_file):
        for section in CONFIGURATION_SECTIONS:
            options.update(dict(configuration_parser.items(section)))
    else:
        options.update(DEFAULT_OPTIONS)

    options.update(vars(arguments))
    return options



"""Contains the code related to options parsing."""

from argparse import (RawDescriptionHelpFormatter,
                      ArgumentParser,
                      SUPPRESS)

from .constants import DEFAULT_CONFIGURATION_FILE


def parse_arguments():
    """Create an :class:`argparse.ArgumentParser` and parse the command-line arguments."""

    parser = ArgumentParser(formatter_class=RawDescriptionHelpFormatter,
                            prog='cmus_notify',
                            description='Display a notification about Cmus\'s '
                                        'current status',
                            epilog='Format String Parameters\n'
                                   '========================\n\n'
                                   'The available arguments to the format '
                                   'strings are the following:\n\n'
                                   '  - album: The song\'s album\n'
                                   '  - artist: The song\'s artist\n'
                                   '  - date: The song\'s release date\n'
                                   '  - discnumber: The song\'s disc\'s number\n'
                                   '  - duration: The song\'s duration\n'
                                   '  - file: The song\'s file\'s path\n'
                                   '  - status: Cmus current status\n'
                                   '  - title: The song\'s title\n'
                                   '  - tracknumber: The song\'s track number')
    parser.add_argument('parse',
                        type=str,
                        metavar='INFORMATION',
                        help='Parse the given information')
    parser.add_argument('-a',
                        '--application_name',
                        default=SUPPRESS,
                        type=str,
                        help='The name of the application')
    parser.add_argument('-b',
                        '--body',
                        required=False,
                        metavar='BODY_FORMAT_STRING',
                        default=SUPPRESS,
                        type=str,
                        help='A format string that can be specified to tell the'
                             ' software how to format the body. The syntax is '
                             'the same as Python\'s. The available options are '
                             'specified at the end of this help message. (i.e. '
                             '\'Artist: {artist}\')')
    parser.add_argument('-t',
                        '--title',
                        required=False,
                        metavar='TITLE_FORMAT_STRING',
                        default=SUPPRESS,
                        type=str,
                        help='A format string that can be specified to tell the'
                             ' software how to format the title. The syntax is '
                             'the same as Python\'s. The available options are '
                             'specified at the end of this help message. (i.e. '
                             '\'Now playing: {title}\')')
    parser.add_argument('-f',
                        '--configuration_file',
                        required=False,
                        default=DEFAULT_CONFIGURATION_FILE,
                        type=str,
                        help='The path to the configuration file. If it is not '
                             'specified, the program will use the default '
                             'values of the other options.')
    parser.add_argument('-c',
                        '--custom_notification',
                        required=False,
                        default=SUPPRESS,
                        type=str,
                        help='The path to a custom implementation of the '
                             'notification class. If it is not specified, the '
                             'standard implementation will be used (the one using '
                             'notify2).')
    return parser.parse_args()



"""Contains the types related to Cmus output."""

from .constants import DEFAULT_STATUS_DISPLAY


class StatusInformation:
    """The status information of Cmus"""

    def __init__(self, **kwargs) -> None:
        """Initialize a :class:`StatusInformation` object.

        :param kwargs: See below.

        :Keyword Arguments:
            * *album* (``str``) --- The name of the song's album
            * *albumartist* (``str``) --- The name of the album's artist
            * *artist* (``str``) --- The name of the song's artist(s)
            * *date* (``str``) --- The date at which the song was composed
            * *discnumber* (``int``) --- The disc's number
            * *duration* (``int``) --- The duration of the song
            * *file* (``str``) --- The place from which the song is read
            * *status* (``str``) --- Cmus current status
            * *title* (``str``) --- The song's title
            * *tracknumber* (``int``) --- The song's number
        """
        self.album = kwargs.get('album', DEFAULT_STATUS_DISPLAY)
        self.albumartist = kwargs.get('albumartist', DEFAULT_STATUS_DISPLAY)
        self.artist = kwargs.get('artist', DEFAULT_STATUS_DISPLAY)
        self.date = kwargs.get('date', DEFAULT_STATUS_DISPLAY)
        self.discnumber = kwargs.get('discnumber', DEFAULT_STATUS_DISPLAY)
        self.duration = kwargs.get('duration', DEFAULT_STATUS_DISPLAY)
        self.file = kwargs.get('file', DEFAULT_STATUS_DISPLAY)
        self.status = kwargs.get('status', DEFAULT_STATUS_DISPLAY)
        self.title = kwargs.get('title', DEFAULT_STATUS_DISPLAY)
        self.tracknumber = kwargs.get('tracknumber', DEFAULT_STATUS_DISPLAY)



"""Contains the main code of the package."""

import importlib
import os
import sys

from .configurations import parse_configuration_file
from .parsers import parse_status_information


def main():
    """Main entry point of the package."""
    arguments = parse_configuration_file()
    information = parse_status_information(arguments['parse'])
    custom_notifier = load_notifier(arguments)
    custom_notifier(arguments, information)


def load_notifier(arguments):
    """Load the custom notification module.

    :param arguments: The parsed arguments
    :returns: The notifier to be used for the notifications
    """
    custom_notifier_path = arguments.get('custom_notification', None)
    if custom_notifier_path:
        sys.path.insert(0, os.path.expanduser(os.path.dirname(custom_notifier_path)))
        notifier_module = importlib.import_module(os.path.basename(custom_notifier_path))
        notifier = notifier_module.send_notification
    else:
        from .notifications import send_notification
        notifier = send_notification

    return notifier



#! /usr/bin/env python3

import os
from setuptools import setup


def get_long_description(file_name: str) -> str:
    """Gets the long description from the specified file's name.

    :param file_name: The file's name
    :type file_name: str
    :return: The content of the file
    :rtype: str
    """
    return open(os.path.join(os.path.dirname(__file__), file_name)).read()


if __name__ == '__main__':
    setup(
        name='cmus-notify',
        version='1.4.0',
        description='A package for displaying Cmus current status in notifications',
        author='Antoine Gagne',
        keywords='utilities application cli hook cmus notifications',
        author_email='antoine.gagne.2@ulaval.ca',
        url='https://github.com/AntoineGagne/cmus-notify',
        packages=['cmus_notify'],
        entry_points={
            'console_scripts': ['cmus_notify = cmus_notify.cmus_notify:main']
        },
        license='MIT',
        data_files=[],
        include_package_data=True,
        long_description=get_long_description('README.rst'),
        setup_requires=['pytest-runner', 'flake8'],
        tests_require=['pytest'],
        test_suite='tests',
        scripts=[],
        classifiers=[
            'License :: OSI Approved :: MIT License',
            'Development Status :: 5 - Production/Stable',
            'Operating System :: POSIX :: Linux',
            'Programming Language :: Python :: 3.5',
            'Topic :: Multimedia :: Sound/Audio'
        ],
        install_requires=['notify2>=0.3']
    )



"""Contains code related to notifications."""

from .constants import (DEFAULT_ICON_PATH,
                        DEFAULT_TIMEOUT,
                        ICONS_BY_STATUS)
from .formatters import format_notification_message


def send_notification(arguments, information):
    """Send the notification to the OS with a Python library.

    :param arguments: The parsed arguments
    :param information: The various song informations
    """
    import notify2

    notify2.init(arguments['application_name'])
    title, text = format_notification_message(information,
                                              title=arguments['title'],
                                              body=arguments['body'])
    notification = notify2.Notification(
        title,
        text,
        ICONS_BY_STATUS.get('icon_path', DEFAULT_ICON_PATH)
    )
    notification.set_urgency(arguments.get('urgency', notify2.URGENCY_LOW))
    notification.timeout = arguments.get('timeout', DEFAULT_TIMEOUT)
    notification.show()
